@@ -1,22 +1,14 @@
-#!/usr/bin/env python3
 from __future__ import annotations
-
 from pathlib import PosixPath, WindowsPath, _NormalAccessor, \
   Path, PurePath, _ignore_error
-from typing import Optional, AsyncIterable, List, Union
-from os import name, stat_result
+from typing import Optional, List, Union, AsyncIterable, \
+  Literal, Final
+from os import stat_result
 from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, \
   S_ISCHR, S_ISFIFO
 import os
 
-
-# try:
-#   from typing import TypeAlias
-
-# except ImportError:
-#   TypeAlias = type
-
-from aiofile import async_open, AIOFile
+from aiofile import AIOFile
 from aiofiles import os as async_os
 from aiofiles.os import wrap as method_as_method_coro, \
   wrap as func_as_corofunc
@@ -24,13 +16,19 @@ from aiofiles.os import wrap as method_as_method_coro, \
 from .selectors import _make_selector
 from .flavours import _async_windows_flavour, _async_posix_flavour
 from .wrap import coro_as_method_coro, func_as_method_coro, to_thread
+from .handle import IterableAIOFile, read_full_file
+from .scandir import EntryWrapper, scandir_async
 
 
-# _PathBase: TypeAlias = \
-#   WindowsPath if name == 'nt' else PosixPath
+DEFAULT_ENCODING: Final[str] = 'utf-8'
+ON_ERRORS: Final[str] = 'ignore'
 
 
-DEFAULT_ENCODING: str = 'utf-8'
+TextMode = \
+  Literal['r', 'w', 'a', 'x', 'r+', 'w+', 'a+', 'x+']
+BinaryMode = \
+  Literal['rb', 'wb', 'ab', 'xb', 'r+b', 'w+b', 'a+b', 'x+b']
+FileMode = Union[TextMode, BinaryMode]
 
 
 getcwd = func_as_corofunc(os.getcwd)
@@ -85,14 +83,9 @@ class _AsyncAccessor(_NormalAccessor):
     except ImportError:
       raise NotImplementedError("Path.group() is unsupported on this system")
 
-  def _scandir_results(self, *args, **kwargs) -> List[str]:
-    return list(os.scandir(*args, **kwargs))
-
-  async def scandir(self, *args, **kwargs) -> AsyncIterable[str]:
-    results = await to_thread(self._scandir_results, *args, **kwargs)
-
-    for result in results:
-      yield result
+  async def scandir(self, *args, **kwargs) -> AsyncIterable[EntryWrapper]:
+    async for entry in scandir_async(*args, **kwargs):
+      yield entry
 
 
 _async_accessor = _AsyncAccessor()
@@ -104,30 +97,31 @@ class AsyncPurePath(PurePath):
 
   def __new__(cls, *args):
     if cls is AsyncPurePath:
-        cls = PureAsyncWindowsPath if os.name == 'nt' else PureAsyncPosixPath
+      cls = PureAsyncWindowsPath if os.name == 'nt' else PureAsyncPosixPath
     return cls._from_parts(args)
 
 
 class PureAsyncPosixPath(AsyncPurePath):
-    """PurePath subclass for non-Windows systems.
-    On a POSIX system, instantiating a PurePath should return this object.
-    However, you can also instantiate it directly on any system.
-    """
-    _flavour = _async_posix_flavour
-    __slots__ = ()
+  """PurePath subclass for non-Windows systems.
+  On a POSIX system, instantiating a PurePath should return this object.
+  However, you can also instantiate it directly on any system.
+  """
+  _flavour = _async_posix_flavour
+  __slots__ = ()
 
 
 class PureAsyncWindowsPath(AsyncPurePath):
-    """PurePath subclass for Windows systems.
-    On a Windows system, instantiating a PurePath should return this object.
-    However, you can also instantiate it directly on any system.
-    """
-    _flavour = _async_windows_flavour
-    __slots__ = ()
+  """PurePath subclass for Windows systems.
+  On a Windows system, instantiating a PurePath should return this object.
+  However, you can also instantiate it directly on any system.
+  """
+  _flavour = _async_windows_flavour
+  __slots__ = ()
 
 
 class AsyncPath(Path, AsyncPurePath):
-  _flavour = _async_windows_flavour if os.name == 'nt' else _async_posix_flavour
+  _flavour = \
+    _async_windows_flavour if os.name == 'nt' else _async_posix_flavour
   _accessor = _async_accessor
 
   def _init(self, template: Optional[AsyncPath] = None):
@@ -135,13 +129,14 @@ class AsyncPath(Path, AsyncPurePath):
 
   def __new__(cls, *args, **kwargs):
     if cls is AsyncPath:
-        cls = AsyncWindowsPath if os.name == 'nt' else AsyncPosixPath
+      cls = AsyncWindowsPath if os.name == 'nt' else AsyncPosixPath
 
     self = cls._from_parts(args, init=False)
 
     if not self._flavour.is_supported:
-        raise NotImplementedError("cannot instantiate %r on your system"
-                                  % (cls.__name__,))
+      name: str = cls.__name__
+      raise NotImplementedError(f"cannot instantiate {name} on your system")
+
     self._init()
     return self
 
@@ -151,32 +146,39 @@ class AsyncPath(Path, AsyncPurePath):
 
   def open(
     self,
-    mode: str = 'r',
+    mode: FileMode = 'r',
     buffering: int = -1,
     encoding: Optional[str] = DEFAULT_ENCODING,
-    errors: Optional[Exception] = None,
-    newline: Optional[str] = None
-  ) -> AIOFile:
-    return AIOFile(
+    errors: Optional[str] = ON_ERRORS,
+    newline: Optional[str] = '\n'
+  ) -> IterableAIOFile:
+    return IterableAIOFile(
       self._path,
       mode,
-      encoding=encoding
+      encoding=encoding,
+      errors=errors,
+      newline=newline,
     )
 
   async def read_text(
     self,
     encoding: Optional[str] = DEFAULT_ENCODING,
-    errors: Optional[Exception] = None
+    errors: Optional[str] = ON_ERRORS
   ) -> str:
-    async with self.open('r', encoding=encoding) as file:
-      return await file.read()
 
-  async def read_bytes(
-    self,
-    encoding: Optional[str] = None,
-    errors: Optional[Exception] = None
-  ) -> bytes:
-    async with self.open(mode='rb') as file:
+    async with self.open('r', encoding=encoding, errors=errors) as file:
+      return await file.read_text()
+
+    #path = str(await self.resolve())
+
+    #return await read_full_file(
+      #path,
+      #encoding=encoding,
+      #errors=errors
+    #)
+
+  async def read_bytes(self) -> bytes:
+    async with self.open('rb') as file:
       return await file.read()
 
   async def write_bytes(self, data: bytes) -> int:
@@ -185,14 +187,15 @@ class AsyncPath(Path, AsyncPurePath):
     """
     # type-check for the buffer interface before truncating the file
     view = memoryview(data)
+
     async with self.open(mode='wb') as f:
-      return await f.write(view)
+      return await f.write(data)
 
   async def write_text(
     self,
     data: str,
-    encoding: Optional[str] = 'utf-8',
-    errors: Optional[Exception] = None,
+    encoding: Optional[str] = DEFAULT_ENCODING,
+    errors: Optional[str] = ON_ERRORS,
     newline: Optional[str] = None
   ) -> int:
     """
@@ -213,7 +216,7 @@ class AsyncPath(Path, AsyncPurePath):
     """
     Return the path to which the symbolic link points.
     """
-    path = await self._accessor.readlink(self)
+    path: str = await self._accessor.readlink(self)
     obj = self._from_parts((path,), init=False)
     obj._init(template=self)
     return obj
@@ -342,24 +345,44 @@ class AsyncPath(Path, AsyncPurePath):
     await self._accessor.symlink(target, self, target_is_directory)
 
   async def exists(self) -> bool:
+    """
+    Whether this path exists.
+    """
     try:
-      async with async_open(self._path, 'rb'):
-        pass
+      await self.stat()
 
-      return True
+    except OSError as e:
+      if not _ignore_error(e):
+        raise
 
-    except IsADirectoryError:
-      return True
-
-    except FileNotFoundError:
       return False
 
+    except ValueError:
+      # Non-encodable path
+      return False
+
+    return True
+
+  #async def exists(self) -> bool:
+    #try:
+      #async with async_open(self._path, 'rb'):
+        #pass
+
+      #return True
+
+    #except IsADirectoryError:
+      #return True
+
+    #except FileNotFoundError:
+      #return False
+
   @classmethod
   async def cwd(cls: type) -> str:
     """Return a new path pointing to the current working directory
     (as returned by os.getcwd()).
     """
-    return cls(await getcwd())
+    cwd: str = await getcwd()
+    return cls(cwd)
 
   @classmethod
   async def home(cls) -> AsyncPath:
@@ -367,7 +390,9 @@ class AsyncPath(Path, AsyncPurePath):
     returned by os.path.expanduser('~')).
     """
     coro = cls()._flavour.gethomedir(None)
-    return cls(await coro)
+    homedir: str = await coro
+
+    return cls(homedir)
 
   async def samefile(self, other_path: Union[AsyncPath, Path]) -> bool:
     """Return whether other_path is the same or not as this file
@@ -382,10 +407,10 @@ class AsyncPath(Path, AsyncPurePath):
 
     else:
       try:
-        other_st = other_path.stat()
+        other_st = await to_thread(other_path.stat)
 
       except AttributeError:
-        other_st = self._accessor.stat(other_path)
+        other_st = await to_thread(self._accessor.stat, other_path)
 
     return os.path.samestat(
       await self.stat(),
@@ -430,7 +455,9 @@ class AsyncPath(Path, AsyncPurePath):
     if drv or root:
       raise NotImplementedError("Non-relative patterns are unsupported")
 
-    selector = _make_selector(("**",) + tuple(pattern_parts), self._flavour)
+    #selector = _make_selector(("**",) + tuple(pattern_parts), self._flavour)
+    parts = ("**", *pattern_parts)
+    selector = _make_selector(parts, self._flavour)
 
     async for p in selector.select_from(self):
       yield p
@@ -456,16 +483,17 @@ class AsyncPath(Path, AsyncPurePath):
     normalizing it (for example turning slashes into backslashes under
     Windows).
     """
-    s = await self._flavour.resolve(self, strict=strict)
+    s: Optional[str] = await self._flavour.resolve(self, strict=strict)
 
     if s is None:
       # No symlink resolution => for consistency, raise an error if
       # the path doesn't exist or is forbidden
       await self.stat()
-      s = str(await self.absolute())
+      path = await self.absolute()
+      s = str(path)
 
     # Now we have no symlinks in the path, it's safe to normalize it.
-    normed = self._flavour.pathmod.normpath(s)
+    normed: str = self._flavour.pathmod.normpath(s)
     obj = self._from_parts((normed,), init=False)
     obj._init(template=self)
     return obj
@@ -664,12 +692,21 @@ class AsyncPath(Path, AsyncPurePath):
 
     return self
 
+  async def iterdir(self) -> AsyncIterable[AsyncPath]:
+    names = await self._accessor.listdir(self)
 
-class AsyncPosixPath(AsyncPath, PureAsyncPosixPath):
+    for name in names:
+      if name in {'.', '..'}:
+        continue
+
+      yield self._make_child_relpath(name)
+
+
+class AsyncPosixPath(PosixPath, AsyncPath, PureAsyncPosixPath):
   __slots__ = ()
 
 
-class AsyncWindowsPath(AsyncPath, PureAsyncWindowsPath):
+class AsyncWindowsPath(WindowsPath, AsyncPath, PureAsyncWindowsPath):
   __slots__ = ()
 
   async def is_mount(self) -> int:
